﻿##############################################################################
#                                                                            #
# Copyright (c) Microsoft Corporation.                                       #
#                                                                            #
# Forefront Protection Server Script Kit (FPSSK)                             #
#                                                                            #
# LICENSED UNDER THE MICROSOFT SOFTWARE LICENSE TERMS PROVIDED WITH THIS     #
# CODE (EULA.RTF). YOUR USE OF THIS CODE IS SUBJECT TO THE TERMS OF THAT     #
# LICENSE. THIS CODE IS PROVIDED *AS IS* WITHOUT WARRANTY OF ANY KIND,       #
# EITHER EXPRESS OR IMPLIED, INCLUDING ANY IMPLIED WARRANTIES OF FITNESS     #
# FOR A PARTICULAR PURPOSE, MERCHANTABILITY, OR NON-INFRINGEMENT.            #
#                                                                            #
##############################################################################

# Globals
$Transcribbing = $null # Signals whether transcribbing is active

# Print help text for the script kit
function help
{
" `
NAME `
    FPSSK.ps1 `
`
SYNOPSIS `
    Forefront Protection Server Script Kit (FPSSK). `
`
SYNTAX `
    FPSSK.ps1 `
        [-Compare] `
            [-XmlPath] <string>          (required) `
            [ [-CsvPath] <string>        (optional) `
              [-ComputerName] <string[]> (optional) `
              [-DiscoverCsv] <string>    (optional) `
              [-Log] <string> ]          (optional) `
        [-Discover] `
            [-Domain] <string>           (required) `
            [ [-ExportCsv] <string> ]    (optional) `
        [-Export] `
            [-Path] <string>             (required) `
            [ [-ComputerName] <string[]> (optional) `
              [-Element] <string[]>      (optional) `
              [-DiscoverCsv] <string>    (optional) `
              [-Log] <string> ]          (optional) `
        [-Import] `
            [-XmlPath] <string>          (required) `
            [ [-CsvPath] <string>        (optional) `
              [-ComputerName] <string[]> (optional) `
              [-DiscoverCsv] <string>    (optional) `
              [-Log] <string>            (optional) `
              [-NoConfirm] ]             (optional) `
        [-Report] `
            [ [-ComputerName] <string[]> (optional) `
              [-DiscoverCsv] <string>    (optional) `
              [-Log] <string> ]          (optional) `
`
DESCRIPTION `
    The Forefront Protection Server Script Kit (FPSSK) enables you to manage `
    the configuration of multiple servers across your deployment from one `
    console. It also provides basic compliance and reporting capabilities to `
    detect configuration drift and monitor server statistics. `
`
PARAMETERS `
    ACTIONS `
        -Compare `
            Compares the Forefront Protection 2010 configuration files `
            obtained by using the Export action with the configuration `
            settings on one or more servers running Forefront Protection `
            2010. `
`
        -Discover `
            Identifies all Forefront Protection 2010 servers in an specified `
            Active Directory domain. `
`
        -Export `
            Captures the Forefront Protection 2010 configuration settings `
            from the specified array of computers and stores the settings in `
            an .xml file and a .csv file on the local computer. The .xml file `
            contains the base configuration settings and the .csv file `
            contains advanced configuration options. You should only modify `
            the advanced configuration settings if you are instructed to do `
            so by Microsoft support personnel. `
`
        -Import `
            Imports the configuration settings captured using the Export `
            action to the specified array of target computers. `
`
        -Report `
            Generates a basic Forefront Protection Server status report for `
            the specified computers. `
`
    OPTIONS `
        -ComputerName <string[]> `
            Specifies a an array of target computers for the selected action. `
            Enter multiple computer names as a comma-separated list of `
            computer names. The default is the local computer. `
`
            Type the NetBIOS name or a fully qualified domain name of one or `
            more remote computers. To specify the local computer, type the `
            computer name, ""localhost"", or a dot (.), or simply do not `
            specify this parameter. When the computer is in a different `
            domain than the user, the fully qualified domain name is `
            required. `
`
        -CsvPath <string> `
            Specifies a full file path to the Forefront Protection 2010 `
            configuration setting .csv file. You should only modify the `
            advanced configuration settings contained in this file if you are `
            instructed to do so by Microsoft support personnel. `
`
        -DiscoverCsv <string> `
            Specifies the full file path to the .csv file that was created `
            using the Discover action and that contains a list of computers `
            running Forefront Protection 2010 in a domain. `
`
        -Domain <string> `
            Specifies the fully qualified name of an Active Directory domain. `
`
        -Element <string[]> `
            Exports only certain configuration items. Enter multiple items as `
            an array. The items include any found in the full .xml file. `
`
        -ExportCsv <string> `
            Specifies a directory path for the .csv file export. `
`
        -Log <string> `
            Specifies the directory path to the location where the action `
            writes log files. This option also tells the action to write the `
            results to a log file. `
`
        -NoConfirm `
            Specifies that no warnings should prompt the user for input. `
`
        -Path <string> `
            Specifies a directory path for the configuration file export. `
`
        -XmlPath <string> `
            Specifies a full file path to the Forefront Protection 2010 `
            configuration settings .xml file. `
`
NOTES `
        This script kit uses the Windows PowerShell 2.0 remoting `
        infrastructure. To use this script kit, the local computer and any `
        remote computers must be configured for Windows PowerShell remoting. `
        For more information, see about_Remote_Requirements. `
`
        In Windows Vista and later versions of Windows, to connect to the `
        local computer using the script kit, you must start Windows `
        PowerShell by selecting the ""Run as administrator"" option. `
`
    -------------------------- EXAMPLE 1 -------------------------- `
`
    C:\PS>.\FPSSK.ps1 -Compare -XmlPath .\FPSSK-Export.xml -DiscoverCsv .\FPSSK-Discover.csv `
`
    -------------------------- EXAMPLE 2 -------------------------- `
`
    C:\PS>.\FPSSK.ps1 -Discover -Domain domain.com -ExportCsv . `
`
    -------------------------- EXAMPLE 3 -------------------------- `
`
    C:\PS>.\FPSSK.ps1 -Export -Path . -ComputerName . `
`
    -------------------------- EXAMPLE 4 -------------------------- `
`
    C:\PS>.\FPSSK.ps1 -Import -XmlPath .\FPSSK-Export.xml -ComputerName server01, server02 -NoConfirm `
`
    -------------------------- EXAMPLE 5 -------------------------- `
`
    C:\PS>.\FPSSK.ps1 -Report -ComputerName server03, server04 -Log . `
`
"
}

# Log all activity printing in PowerShell to a local file
function Transcribe ($log, $FileName)
{
    if ($log -ne $null)
    {
        $date = get-date -uformat "%Y%m%d%H%M%S"

        # Generate a full path for the log file
        $logPath = "$log\FPSSK-$FileName-$date.txt"

        # Start logging
        $result =
            Start-Transcript `
                -Path $logPath `
                -ErrorAction:SilentlyContinue `
                -ErrorVariable Global:errorString
        if ($result -eq $null)
        {
            printError "Error: Logging failed to $logPath!"
            return
        }

        $Global:transcribbing = $true
    }
}

# Stop any existing logging
function TranscribeStop
{
    if ($Global:transcribbing)
    {
        Stop-Transcript -ErrorAction:SilentlyContinue
        $Global:transcribbing = $false
    }
}

# Load Forefront's CmdLet for Exchange or Sharepoint (will autodetect either)
function loadFSS
{
    # This is expected to fail if the Forefront PS Snapin isn't loaded
    invokeCommand "Get-PSSnapin fsspssnapin"
    if ($Global:errorString.value -ne $Null)
    {
        # Load the snapin module
        invokeCommand "Add-PSSnapin fsspssnapin"
        if ($Global:errorString.value -ne $Null)
        {
            return
        }
    }

    # Attempt to run a Exchange and Sharepoint command, assume that a
    # computer CANNOT have both at the same time
    $detectFse = invokeCommand "Get-FseProductInfo"
    $detectFssp = invokeCommand "Get-FsspProductInfo"
    if ($detectFse -ne $Null)
    {
        $Global:FSEPresent = $true
        $Global:extension = "FSE"
    }
    elseif ($detectFssp -ne $Null)
    {
        $Global:FSEPresent = $false
        $Global:extension = "FSSP"
    }
}

######################### Main function, so to speak #########################

    # Init
    $Global:transcribbing = $compare = $discover = $export = $import =
        $report = $actionSelected = $quiet = $false

    $domain = $exportcsv = $element = $path = $xmlPath = $csvPath =
        $computername = $discoverCsv = $log = $null

    $actionError = "Error: Only one action at a time may be performed!"
    $arrayError = "Error: This parameter does not support a string array!"
    $computerError = "Error: You may not specify -ComputerName AND -DiscoverCsv"
    $fileError = "Error: Path must be an existing file!"
    $nullError = "Error: One or more parameters is NULL!"
    $pathError = "Error: Path must be an existing folder!"

    # Find out where we are running
    if ($myinvocation.mycommand.path -ne $null)
    {
        $currentDir = split-path $myinvocation.mycommand.path
    }
    else
    {
        # Sometimes $myinvocation is null, it depends on the PS console host
        $currentDir = "."
    }

    # Load all of the scripts globally so we can call their functions
    . $currentDir\FPSSK-Common.ps1
    . $currentDir\FPSSK-Compare.ps1
    . $currentDir\FPSSK-Discover.ps1
    . $currentDir\FPSSK-Export.ps1
    . $currentDir\FPSSK-Import.ps1
    . $currentDir\FPSSK-Report.ps1

    # Make sure they were actually loaded
    if (($Global:ComparePs1Loaded -ne $true)  -or
        ($Global:CommonPs1Loaded -ne $true)   -or
        ($Global:DiscoverPs1Loaded -ne $true) -or
        ($Global:ExportPs1Loaded -ne $true)   -or
        ($Global:ImportPs1Loaded -ne $true)   -or
        ($Global:ReportPs1Loaded -ne $true))
    {
        $errorMessage = "Error: Failed to load one or more dependent scripts!"
        write-host -foregroundcolor red "`n$errorMessage`n`n"
        return
    }

    # Process command-line
    for ($i = 0; $i -lt $args.count; $i++)
    {
        switch ($args[$i])
        {
            ##### Compare command #####
            "-Compare"
            {
                if (!$actionSelected)
                {
                    $compare = $true
                    $actionSelected = $true
                }
                else
                {
                    help
                    printError $actionError
                    return
                }
                break
            }
            ##### Discover command #####
            "-Discover"
            {
                if (!$actionSelected)
                {
                    $discover = $true
                    $actionSelected = $true
                }
                else
                {
                    help
                    printError $actionError
                    return
                }
                break
            }
            "-Domain"
            {
                $domain = $args[++$i]
                if ($domain -eq $null)
                {
                    help
                    printError $nullError
                    return
                }
                $arrayCount = Get-Member -InputObject $domain -Name count
                if ($arrayCount -ne $null)
                {
                    help
                    printError "$arrayError ($domain)"
                    return
                }
                break
            }
            "-ExportCsv"
            {
                $exportcsv = $args[++$i]
                if ($exportcsv -eq $null)
                {
                    help
                    printError $nullError
                    return
                }
                $arrayCount = Get-Member -InputObject $exportcsv -Name count
                if ($arrayCount -ne $null)
                {
                    help
                    printError "$arrayError ($exportcsv)"
                    return
                }

                $success =
                    Test-Path -LiteralPath $exportcsv -PathType Container
                if (!$success)
                {
                    help
                    printError "$pathError ($exportcsv)"
                    return
                }
                break
            }
            ##### Export command #####
            "-Export"
            {
                if (!$actionSelected)
                {
                    $export = $true
                    $actionSelected = $true
                }
                else
                {
                    help
                    printError $actionError
                    return
                }
                break
            }
            "-Element"
            {
                $element = $args[++$i]
                if ($element -eq $null)
                {
                    help
                    printError $nullError
                    return
                }
                break
            }
            "-Path"
            {
                $path = $args[++$i]
                if ($path -eq $null)
                {
                    help
                    printError $nullError
                    return
                }
                $arrayCount = Get-Member -InputObject $path -Name count
                if ($arrayCount -ne $null)
                {
                    help
                    printError "$arrayError ($path)"
                    return
                }

                $success =
                    Test-Path -LiteralPath $path -PathType Container
                if (!$success)
                {
                    help
                    printError "$pathError ($path)"
                    return
                }
                break
            }
            ##### Import command #####
            "-Import"
            {
                if (!$actionSelected)
                {
                    $import = $true
                    $actionSelected = $true
                }
                else
                {
                    help
                    printError $actionError
                    return
                }
                break
            }
            "-XmlPath"
            {
                $xmlPath = $args[++$i]
                if ($xmlPath -eq $null)
                {
                    help
                    printError $nullError
                    return
                }
                $arrayCount = Get-Member -InputObject $xmlPath -Name count
                if ($arrayCount -ne $null)
                {
                    help
                    printError "$arrayError ($xmlPath)"
                    return
                }

                $success =
                    Test-Path -LiteralPath $xmlPath -PathType Leaf
                if (!$success)
                {
                    help
                    printError "$fileError ($xmlPath)"
                    return
                }
                break
            }
            "-CsvPath"
            {
                $csvPath = $args[++$i]
                if ($csvPath -eq $null)
                {
                    help
                    printError $nullError
                    return
                }
                $arrayCount = Get-Member -InputObject $csvPath -Name count
                if ($arrayCount -ne $null)
                {
                    help
                    printError "$arrayError ($csvPath)"
                    return
                }

                $success =
                    Test-Path -LiteralPath $csvPath -PathType Leaf
                if (!$success)
                {
                    help
                    printError "$fileError ($csvPath)"
                    return
                }
                break
            }
            ##### Report command #####
            "-Report"
            {
                if (!$actionSelected)
                {
                    $report = $true
                    $actionSelected = $true
                }
                else
                {
                    help
                    printError $actionError
                    return
                }
                break
            }
            "-ComputerName"
            {
                if ($discoverCsv -ne $null)
                {
                    help
                    printError $computerError
                    return
                }
                $computername = $args[++$i]
                if ($computername -eq $null)
                {
                    help
                    printError $nullError
                    return
                }
                break
            }
            "-DiscoverCsv"
            {
                if ($computername -ne $null)
                {
                    help
                    printError $computerError
                    return
                }
                $discoverCsv = $args[++$i]
                if ($discoverCsv -eq $null)
                {
                    help
                    printError $nullError
                    return
                }
                $arrayCount = Get-Member -InputObject $discoverCsv -Name count
                if ($arrayCount -ne $null)
                {
                    help
                    printError "$arrayError ($discoverCsv)"
                    return
                }

                $success =
                    Test-Path -LiteralPath $discoverCsv -PathType Leaf
                if (!$success)
                {
                    help
                    printError "$fileError ($discoverCsv)"
                    return
                }
                break
            }
            "-Log"
            {
                $log = $args[++$i]
                if ($log -eq $null)
                {
                    help
                    printError $nullError
                    return
                }
                $arrayCount = Get-Member -InputObject $log -Name count
                if ($arrayCount -ne $null)
                {
                    help
                    printError "$arrayError ($log)"
                    return
                }

                $success =
                    Test-Path -LiteralPath $log -PathType Container
                if (!$success)
                {
                    help
                    printError "$pathError ($log)"
                    return
                }
                break
            }
            "-NoConfirm"
            {
                $quiet = $true
            }
            default
            {
                help
                return
            }
        }
    }

    # Handle the easy stuff first. Discover doesn't require a remote computer
    if ($discover)
    {
        if ($domain -eq $null)
        {
            help
            printError "Error: Discovery requires a domain name!"
            return
        }

        # Call the discover script to do the work (export to csv is optional)
        cmdDiscover $domain $exportcsv
        return
    }
    elseif (!$compare -and !$export -and !$import -and !$report)
    {
        # User didn't supply any action to perform
        help
        return
    }

    # Process params for the rest of the actions
    if ($export -and ($path -eq $null))
    {
        help
        printError "Error: Export requires a path!"
        return
    }
    elseif (($compare -or $import) -and ($xmlPath -eq $null))
    {
        help
        printError "Error: The XML file path is required!"
        return
    }

    # If the user provided a CSV full of computer names, grab the list
    if ($discoverCsv -ne $null)
    {
        # Pipe the computer names from the CSV
        $computername =
            Import-Csv `
                -Path $discoverCsv `
                -ErrorAction:SilentlyContinue |
                    %{$_.Computer}
        if ($computername -eq $null)
        {
            help
            printError "Error: Import failed from $discoverCsv!"
            return
        }
    }
    # Default to the local machine
    elseif ($computername -eq $null)
    {
        $computername = "localhost"
    }
    
    # Display import warning as needed
    if ($import)
    {
        $warning =
            "Warning: When importing configuration settings, all of the " +
            "settings on the target computer will be overwritten. It is " +
            "recommended that you only import configuration settings onto " +
            "computers with similar role configurations."
        printWarning $warning

        if (!$quiet)
        {
            $confim = Read-Host -Prompt "Continue? Y/[N] "
            if ($confim[0] -ne "y")
            {
                return
            }
        }
    }

    # Sort and filter out any duplicates in the computernames
    $computername = $computername | Sort-Object | Get-Unique

    # Walk through all computers and perform the desired action
    foreach ($computer in $computername)
    {
        # Clean up any existing session
        if ($Global:server -ne $null)
        {
            Remove-PSSession `
                -Session $Global:server `
                -ErrorAction:SilentlyContinue

            $Global:server = $null
        }

        # If we are transcribing, stop
        TranscribeStop

        # There is a delay while connection a session, notify user
        write-host "`nConnecting to $computer`n"

        # Create a new PS session with the computer
        $Global:server = New-PSSession `
                            -ComputerName $computer `
                            -ErrorAction:SilentlyContinue
        if ($Global:server -eq $null)
        {
            printError "Error: Failed to connect to $computer"
            continue
        }

        initRemoteError

        # Load the Forefront snapin on the remote computer
        loadFSS
        if ($Global:extension -eq $null)
        {
            $errorMessage = "Error: Unable to find the Forefront " +
                            "Server Protection Powershell Interface on " +
                            "$computer"
            printError $errorMessage
            continue
        }

        # Call the script to do the actual work
        if ($compare)
        {
            # If the user wants a log, start transcribing
            Transcribe $log "Compare-$computer"

            cmdCompare $xmlPath $csvPath
        }
        elseif ($export)
        {
            # If the user wants a log, start transcribing
            Transcribe $log "Export-$computer"

            $ignored = cmdExport $path $false $element
        }
        elseif ($import)
        {
            # If the user wants a log, start transcribing
            Transcribe $log "Import-$computer"

            cmdImport $xmlPath $csvPath
        }
        elseif ($report)
        {
            # If the user wants a log, start transcribing
            Transcribe $log "Report-$computer"

            cmdReport
        }
    } # Foreach end

    # Clean up any existing session
    if ($Global:server -ne $null)
    {
        Remove-PSSession `
            -Session $Global:server `
            -ErrorAction:SilentlyContinue

        $Global:server = $null
    }

    # If we are still transcribbing, stop
    TranscribeStop

    # If a report was running with more than one computer, print a summary
    $arrayCount = Get-Member -InputObject $ComputerName -Name count
    if ($report -and $arrayCount -ne $null)
    {
        Transcribe $log "Report-Summary"
        cmdReportSummary
        TranscribeStop
    }

# SIG # Begin signature block
# MIIbKAYJKoZIhvcNAQcCoIIbGTCCGxUCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUgl9/CEJWM7228ywv603FvN2X
# SmGgghXnMIIEhTCCA22gAwIBAgIKYQX3HgAAAAAAMjANBgkqhkiG9w0BAQUFADB5
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSMwIQYDVQQDExpN
# aWNyb3NvZnQgQ29kZSBTaWduaW5nIFBDQTAeFw0wOTA3MTMyMzAwMThaFw0xMDEw
# MTMyMzEwMThaMIGDMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQ
# MA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9u
# MQ0wCwYDVQQLEwRNT1BSMR4wHAYDVQQDExVNaWNyb3NvZnQgQ29ycG9yYXRpb24w
# ggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC1jKmclIRhGswuIsSgk2QJ
# 3oWhsJBd/MlsaLXx7/yOa/T4oG7XUd5goydj1PbftkzSRQunPdhv/u2nGGPTH8aB
# 1L8HN5dAK3/nX1o12ZNPq/MqkvCmjYHsBSTc27hhY1NhAjU41rj17ntUpslYNen0
# Fe/aB2wFLU2ktrqbO7gxMH5FvKmhJTkMQ6RHO5ecIERyN89QoE1N9MNl3ETRCDs4
# 3T7FF2xGx61zJMCPicHq6rrNaKkS1lgMUxYM/jkD1oHyPlxX83rbZuEHGNkZAxT2
# 5kWZKlFMkNda76+3OwwpTW0grge3ki7oacqfMELnwzKKCyJKMuicDMZRAyYBbZJz
# AgMBAAGjggECMIH/MBMGA1UdJQQMMAoGCCsGAQUFBwMDMB0GA1UdDgQWBBSHgbff
# 7vp3+lkFYzcECtfp2qDg4DAOBgNVHQ8BAf8EBAMCB4AwHwYDVR0jBBgwFoAUV0V0
# HF2w9shDBeCMVC2PMqf+SJYwSQYDVR0fBEIwQDA+oDygOoY4aHR0cDovL2NybC5t
# aWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMvQ29kZVNpZ1BDQS5jcmwwTQYI
# KwYBBQUHAQEEQTA/MD0GCCsGAQUFBzAChjFodHRwOi8vd3d3Lm1pY3Jvc29mdC5j
# b20vcGtpL2NlcnRzL0NvZGVTaWdQQ0EuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQA2
# IYT1j4MhTYeoP4vnxaQPTDBuWXSbC5d0Fqmx83Ckr8IMMyTzoEr6weQhxrbZ6lxb
# kWH3Im3/MfwXTtfNQOOd0uW00vqNbAfad5rKK+4NWUfPBMIQZ/rq6UM7t+bRnNmO
# I2BuEpvD2X5cVzLdlA3ffcVIkk1IdreiwFOVPvw6o5N88oUEx8aFOJc2+BKTIgCZ
# TnmvlutaIIOs0FxaMLAA4ms5jySJ2j0MpEZ5YjZTatfJzZII+4flVPV8WcXEd32w
# wSyYx5dsQ6f4IPiDVGSeYPjiSqVn1Em/VIh9UkTzUcAm0cV88InWEQkowkVFgB2Q
# Ir2OogU+PLNPvdc54IiTMIIEyjCCA7KgAwIBAgIKYQSz9QAAAAAADTANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMDgwNzI1MTkxMzQ1
# WhcNMTEwNzI1MTkyMzQ1WjCBszELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
# bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
# b3JhdGlvbjENMAsGA1UECxMETU9QUjEnMCUGA1UECxMebkNpcGhlciBEU0UgRVNO
# OjlFNzgtODY0Qi0wMzlEMSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBT
# ZXJ2aWNlMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAq0YvvKgU8vhG
# PwwFoaRjRQV/mukzcswn0zLsmMPgNIiFzqESl65IUE5IIH9w5Fkz/JlbuQeASCi8
# jbj6ZgJmtfF/HFCWKJNVjYFVqnBbpcnO7y7i5yXDdm8pC58syFOAmsEN472nRb8x
# 3+aGi32JAYsceFg4st2TSlueHdecxvDETdC/J0oHlARxAogNcOb4KucOYeOYS9Hd
# q1S2N/vGRv2kHLz/r+yyl4DcuJP9uCKonSVZcD5cexxBKf5zmSJJgddVJsDs8k4o
# VGI7cC1hkx1Q1Z8BwgXDVRpKh6QAhlSge2Ky6g/6XVzFKhGCl76GGx2rt6sKKfzM
# zTSJv0C2TwIDAQABo4IBGTCCARUwHQYDVR0OBBYEFKCFMV6Hrg+qCgTr4cDTZL08
# u2n1MB8GA1UdIwQYMBaAFCM0+NlSRnAK7UD7dvuzK7DDNbMPMFQGA1UdHwRNMEsw
# SaBHoEWGQ2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3Rz
# L01pY3Jvc29mdFRpbWVTdGFtcFBDQS5jcmwwWAYIKwYBBQUHAQEETDBKMEgGCCsG
# AQUFBzAChjxodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jv
# c29mdFRpbWVTdGFtcFBDQS5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDgYDVR0P
# AQH/BAQDAgbAMA0GCSqGSIb3DQEBBQUAA4IBAQBHcU+2SEXACOucL9MlW9tnURGS
# ptnZshSaJsL9BsHM14j1pjW1zSu2a1hqzfmd81rrRi+h+qcffXpCsrDRGgkoVGxv
# mYs4/DrkwSSqckwSBC/x1Hp2O8XPcScLbfw4Z6fFImIre3/BdqynxSZMqa+Y3Yo4
# Xc9lpIEuqe9WAky5CL5lxB6LIS1up/J1ZLTcGmvQ8SXlwueARDnULwp5fgEqp3hb
# o2tbzgnvOTo5t8RnU7mq82H/nXnjz0hn/6vIxitiCZTQH6J493NJO1hTwQmrTAYL
# 1UAofvaxZsW/UsS+Pjvg+Y50jeYiDMHGM5i1eegM9wB86rJOzICaFi3IpBAFMIIG
# BzCCA++gAwIBAgIKYRZoNAAAAAAAHDANBgkqhkiG9w0BAQUFADBfMRMwEQYKCZIm
# iZPyLGQBGRYDY29tMRkwFwYKCZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQD
# EyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRob3JpdHkwHhcNMDcwNDAz
# MTI1MzA5WhcNMjEwNDAzMTMwMzA5WjB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSEwHwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0Ew
# ggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCfoWyx39tIkip8ay4Z4b3i
# 48WZUSNQrc7dGE4kD+7Rp9FMrXQwIBHrB9VUlRVJlBtCkq6YXDAm2gBr6Hu97IkH
# D/cOBJjwicwfyzMkh53y9GccLPx754gd6udOo6HBI1PKjfpFzwnQXq/QsEIEovmm
# bJNn1yjcRlOwhtDlKEYuJ6yGT1VSDOQDLPtqkJAwbofzWTCd+n7Wl7PoIZd++NIT
# 8wi3U21StEWQn0gASkdmEScpZqiX5NMGgUqi+YSnEUcUCYKfhO1VeP4Bmh1QCIUA
# EDBG7bfeI0a7xC1Un68eeEExd8yb3zuDk6FhArUdDbH895uyAc4iS1T/+QXDwiAL
# AgMBAAGjggGrMIIBpzAPBgNVHRMBAf8EBTADAQH/MB0GA1UdDgQWBBQjNPjZUkZw
# Cu1A+3b7syuwwzWzDzALBgNVHQ8EBAMCAYYwEAYJKwYBBAGCNxUBBAMCAQAwgZgG
# A1UdIwSBkDCBjYAUDqyCYEBWJ5flJRP8KuEKU5VZ5KShY6RhMF8xEzARBgoJkiaJ
# k/IsZAEZFgNjb20xGTAXBgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMT
# JE1pY3Jvc29mdCBSb290IENlcnRpZmljYXRlIEF1dGhvcml0eYIQea0WoUqgpa1M
# c1j0BxMuZTBQBgNVHR8ESTBHMEWgQ6BBhj9odHRwOi8vY3JsLm1pY3Jvc29mdC5j
# b20vcGtpL2NybC9wcm9kdWN0cy9taWNyb3NvZnRyb290Y2VydC5jcmwwVAYIKwYB
# BQUHAQEESDBGMEQGCCsGAQUFBzAChjhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20v
# cGtpL2NlcnRzL01pY3Jvc29mdFJvb3RDZXJ0LmNydDATBgNVHSUEDDAKBggrBgEF
# BQcDCDANBgkqhkiG9w0BAQUFAAOCAgEAEJeKw1wDRDbd6bStd9vOeVFNAbEudHFb
# bQwTq86+e4+4LtQSooxtYrhXAstOIBNQmd16QOJXu69YmhzhHQGGrLt48ovQ7DsB
# 7uK+jwoFyI1I4vBTFd1Pq5Lk541q1YDB5pTyBi+FA+mRKiQicPv2/OR4mS4N9wfi
# cLwYTp2OawpylbihOZxnLcVRDupiXD8WmIsgP+IHGjL5zDFKdjE9K3ILyOpwPf+F
# ChPfwgphjvDXuBfrTot/xTUrXqO/67x9C0J71FNyIe4wyrt4ZVxbARcKFA7S2hSY
# 9Ty5ZlizLS/n+YWGzFFW6J1wlGysOUzU9nm/qhh6YinvopspNAZ3GmLJPR5tH4Lw
# C8csu89Ds+X57H2146SodDW4TsVxIxImdgs8UoxxWkZDFLyzs7BNZ8ifQv+AeSGA
# nhUwZuhCEl4ayJ4iIdBD6Svpu/RIzCzU2DKATCYqSCRfWupW76bemZ3KOm+9gSd0
# BhHudiG/m4LBJ1S2sWo9iaF2YbRuoROmv6pH8BJv/YoybLL+31HIjCPJZr2dHYcS
# ZAI9La9Zj7jkIeW1sMpjtHhUBdRBLlCslLCleKuzoJZ1GtmShxN1Ii8yqAhuoFuM
# Jb+g74TKIdbrHk/Jmu5J4PcBZW+JC33Iacjmbuqnl84xKf8OxVtc2E0bodj6L54/
# LlUWa8kTo/0wggaBMIIEaaADAgECAgphFQgnAAAAAAAMMA0GCSqGSIb3DQEBBQUA
# MF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAXBgoJkiaJk/IsZAEZFgltaWNyb3Nv
# ZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290IENlcnRpZmljYXRlIEF1dGhvcml0
# eTAeFw0wNjAxMjUyMzIyMzJaFw0xNzAxMjUyMzMyMzJaMHkxCzAJBgNVBAYTAlVT
# MRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQK
# ExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2Rl
# IFNpZ25pbmcgUENBMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAn43f
# hTeMsQZWZjZO1ArrNiORHq+rjVjpxM/BnzoKJMTExF6w7hUUxfo+mTNrGWly9HwF
# X+WZJUTXNRmKkNwojpAM79WQYa3e3BhwLYPJb6+FLPjdubkw/XF4HIP9yKm5gmcN
# erjBCcK8FpdXPxyY02nXMJCQkI0wH9gm1J57iNniCe2XSUXrBFKBdXu4tSK4Lla7
# 18+pTjwKg6KoOsWttgEOas8itCMfbNUn57d+wbTVMq15JRxChuKdhfRX2htZLy0m
# kinFs9eFo55gWpTme5x7XoI0S23/1O4n0KLc0ZAMzn0OFXyIrDTHwGyYhErJRHlo
# KN8igw24iixIYeL+EQIDAQABo4ICIzCCAh8wEAYJKwYBBAGCNxUBBAMCAQAwHQYD
# VR0OBBYEFFdFdBxdsPbIQwXgjFQtjzKn/kiWMAsGA1UdDwQEAwIBxjAPBgNVHRMB
# Af8EBTADAQH/MIGYBgNVHSMEgZAwgY2AFA6sgmBAVieX5SUT/CrhClOVWeSkoWOk
# YTBfMRMwEQYKCZImiZPyLGQBGRYDY29tMRkwFwYKCZImiZPyLGQBGRYJbWljcm9z
# b2Z0MS0wKwYDVQQDEyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRob3Jp
# dHmCEHmtFqFKoKWtTHNY9AcTLmUwUAYDVR0fBEkwRzBFoEOgQYY/aHR0cDovL2Ny
# bC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMvbWljcm9zb2Z0cm9vdGNl
# cnQuY3JsMFQGCCsGAQUFBwEBBEgwRjBEBggrBgEFBQcwAoY4aHR0cDovL3d3dy5t
# aWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNyb3NvZnRSb290Q2VydC5jcnQwdgYD
# VR0gBG8wbTBrBgkrBgEEAYI3FS8wXjBcBggrBgEFBQcCAjBQHk4AQwBvAHAAeQBy
# AGkAZwBoAHQAIACpACAAMgAwADAANgAgAE0AaQBjAHIAbwBzAG8AZgB0ACAAQwBv
# AHIAcABvAHIAYQB0AGkAbwBuAC4wEwYDVR0lBAwwCgYIKwYBBQUHAwMwDQYJKoZI
# hvcNAQEFBQADggIBADC8sCCkYqCn7zkmYT3crMaZ0IbELvWDMmVeIj6b1ob46Laf
# yovWO3ULoZE+TN1kdIxJ8oiMGGds/hVmRrg6RkKXyJE31CSx56zT6kEUg3fTyU8F
# X6MUUr+WpC8+VlsQdc5Tw84FVGm0ZckkpQ/hJbgauU3lArlQHk+zmAwdlQLuIlmt
# IssFdAsERXsEWeDYD7PrTPhg3cJ4ntG6n2v38+5+RBFA0r26m0sWCG6kvlXkpjgS
# o0j0HFV6iiDRff6R25SPL8J7a6ZkhU+j5Sw0KV0Lv/XHOC/EIMRWMfZpzoX4CpHs
# 0NauujgFDOtuT0ycAymqovwYoCkMDVxcViNX2hyWDcgmNsFEy+Xh5m+J54/pmLVz
# 03jj7aMBPHTlXrxs9iGJZwXsl521sf2vpulypcM04S+f+fRqOeItBIJb/NCcrnyd
# EfnmtVMZdLo5SjnrfUKzSjs3PcJKeyeY5+JOmxtKVDhqIze+ardI7upCDUkkkY63
# BC6Xb+TnRbuPTf1g2ddZwtiA1mA0e7ehkyD+gbiqpVwJ6YoNvihNftfoD+1leNEx
# X7lm299C5wvMAgeN3/8gBqNFZbSzMo0ukeJNtKnJ+rxrBA6yn+qf3qTJCpb0jffY
# mKjwhQIIWaQgpiwLGvJSBu1p5WQYG+Cjq97KfBRhQ7hl9TajVRMrZyxNGzBMMYIE
# qzCCBKcCAQEwgYcweTELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24x
# EDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
# bjEjMCEGA1UEAxMaTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0ECCmEF9x4AAAAA
# ADIwCQYFKw4DAhoFAKCB2DAZBgkqhkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAcBgor
# BgEEAYI3AgELMQ4wDAYKKwYBBAGCNwIBFTAjBgkqhkiG9w0BCQQxFgQUIFIMM/3k
# T1yhEUvUtgoobCFhiKUweAYKKwYBBAGCNwIBDDFqMGigToBMAEYAbwByAGUAZgBy
# AG8AbgB0ACAAUAByAG8AdABlAGMAdABpAG8AbgAgAFMAZQByAHYAZQByACAAUwBj
# AHIAaQBwAHQAIABLAGkAdKEWgBRodHRwOi8vbWljcm9zb2Z0LmNvbTANBgkqhkiG
# 9w0BAQEFAASCAQAtBbNmvh+zcXmPd4Rq9R0Nn8dfG29hq2GQI53b+amtlR5ZkO8a
# FkRHZ0M45RkhnDGH0s9sQ57rj5OF/SAQPOlPxJfviKyYCbZmFGJvVLMk4BoJYvSn
# QuVcisOyyI7fLwFFh3+BbePPdrqwAaQgBclsYJUogYK/70LMKkVwt1UDrKZRqzqV
# aocbRTyf9BnRJQaLUbYwoIykrZR3Kfa8TE3J4u4fsDYk0cud2eidcBzSAgQSKE01
# BPuKdVCndyoEAXrwAsdxSW/5bD89WlhBXrBtPbAO9vnxQC+9mkqzfrUTY3Yg4IVn
# 6X/EJ0F5p6cKN33iXizbcKQaHgNSzkZQSDQIoYICHTCCAhkGCSqGSIb3DQEJBjGC
# AgowggIGAgEBMIGFMHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9u
# MRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRp
# b24xITAfBgNVBAMTGE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQQIKYQSz9QAAAAAA
# DTAHBgUrDgMCGqBdMBgGCSqGSIb3DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcN
# AQkFMQ8XDTEwMDcyNzIyNTUxNlowIwYJKoZIhvcNAQkEMRYEFNhTcxrhYzi6MMrz
# NY+CfqZ9h4nYMA0GCSqGSIb3DQEBBQUABIIBAIb8cedvRJCh4hSpXooGk431FkUn
# KBsqLpzr9QH5HwRhmE1DB1jZ1RMvL494aYasbw9SN2b1EXFVoeYtL0XXOROPQoCT
# PNyGBfYlbQxtG0dBWwoS3P6bDC1XgImLTI3MgEAalL7O+0s5+pS+jstjWdtjJG9k
# /FuLV63RMrHYq0AZ/VQS5Y5HxCvL8HBR7KRhEUsxfTxrqpOl85GO4oi75I35HV8M
# Nubd9uZQWxsQR7iMSfbkBsGVCCyQEtOUN/wN9cZxHaEn+Lw3hXCIiJZ4GHPnPYel
# rrZDl15q/13kZGvE246h/5jY8Tno1C8DsLs/XRxwMPwVyfDsVVpJZop0BWw=
# SIG # End signature block
